from __future__ import division

import copy

import numpy as np
import tigre
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator
from tigre.utilities.im_3d_denoise import im3ddenoise



class FISTA(IterativeReconAlg):
    """
    Solves the reconstruction problem
    using the projection data PROJ taken over ALPHA angles, correspond-
    ing to the geometry descrived in GEO, using NITER iterations.

    Parameters
    ----------
    :param proj: (np.ndarray, dtype=np.float32)
    Input data, shape = (geo.nDector, nangles)

    :param geo: (tigre.geometry)
    Geometry of detector and image (see examples/Demo code)

    :param angles: (np.ndarray , dtype=np.float32)
    angles of projection, shape = (nangles,3)

    :param niter: (int)
    number of iterations for reconstruction algorithm

    :param kwargs: (dict)
    optional parameters

    Keyword Arguments
    -----------------
    :keyword hyper: (np.float64)
        hyper parameter proportional to the largest eigenvalue of the
        matrix A in the equations Ax-b and ATb.
        Empirical tests show, for the headphantom object:

        nVoxel = np.array([64,64,64]),      hyper (approx=) 2.e8
        nVoxel = np.array([512,512,512]),   hyper (approx=) 2.e4

    :keyword init: (str)
        Describes different initialization techniques.
              "none"     : Initializes the image to zeros (default)
              "FDK"      : intializes image to FDK reconstrucition
              
    :keyword verbose:  (Boolean)
        Feedback print statements for algorithm progress
        default=True

    :keyword OrderStrategy : (str)
        Chooses the subset ordering strategy. Options are:
                 "ordered"        : uses them in the input order, but
                                    divided
                 "random"         : orders them randomly

    :keyword tviter: (int)
        Number of iterations of im3ddenoise for every iteration.
        Default: 20

    :keyword tvlambda: (float)
        Multiplier for lambdaForTV which is proportional to L (hyper)
        Default: 0.1      
        
    :keyword fista_p: (float)
        Default: 1 for standard FISTA 
        0.01 < fista_p <= 0.1 for faster FISTA
        
    :keyword fista_q: (float)
        Default: 1 for standard FISTA 
        0.0 < fista_q <= 1.0 for faster FISTA

    Usage
    --------
    >>> import numpy as np
    >>> import tigre
    >>> import tigre.algorithms as algs
    >>> from tigre.demos.Test_data import data_loader
    >>> geo = tigre.geometry(mode='cone',default_geo=True,
    >>>                         nVoxel=np.array([512,512,512]))
    >>> angles = np.linspace(0,2*np.pi,100)
    >>> src_img = data_loader.load_head_phantom(geo.nVoxel)
    >>> proj = tigre.Ax(src_img,geo,angles)
    >>> output = algs.fista(proj,geo,angles,niter=50
    >>>                                 hyper=2.e4)

    tigre.demos.run() to launch ipython notebook file with examples.


    --------------------------------------------------------------------
    This file is part of the TIGRE Toolbox

    Copyright (c) 2015, University of Bath and
                        CERN-European Organization for Nuclear Research
                        All rights reserved.

    License:            Open Source under BSD.
                        See the full license at
                        https://github.com/CERN/TIGRE/license.txt

    Contact:            tigre.toolbox@gmail.com
    Codes:              https://github.com/CERN/TIGRE/
    --------------------------------------------------------------------
    Coded by:          MATLAB (original code): Ander Biguri
                       PYTHON : Reuben Lindroos

    """

    def __init__(self, proj, geo, angles, niter, **kwargs):
        # Don't precompute W and V
        kwargs.update({"W": None, "V": None})
        kwargs.update(dict(blocksize=angles.shape[0]))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)
        self.lmbda = 0.1
        self.__L__ = 2.0e8 if "hyper" not in kwargs else kwargs["hyper"]
        self.__numiter_tv__ = 20 if "tviter" not in kwargs else kwargs["tviter"]
        self.__lambda__ = 0.1 if "tvlambda" not in kwargs else kwargs["tvlambda"]
        self.__t__ = 1
        self.__bm__ = 1.0 / self.__L__
        self.__p__ = 1 if "fista_p" not in kwargs else kwargs["fista_p"]
        self.__q__ = 1 if "fista_q" not in kwargs else kwargs["fista_q"]

    # overide update_image from iterative recon alg to remove W.
    def update_image(self, geo, angle, iteration):
        """
        VERBOSE:
         for j in range(angleblocks):
             angle = np.array([alpha[j]], dtype=np.float32)
             proj_err = proj[angle_index[j]] - Ax(res, geo, angle, 'Siddon')
             backprj = Atb(proj_err, geo, angle, 'FDK')
             res += backprj
             res[res<0]=0

        :return: None
        """
        self.res += (
            self.__bm__
            * 2
            * tigre.Atb(
                (
                    self.proj[self.angle_index[iteration]]
                    - tigre.Ax(self.res, geo, angle, "interpolated", gpuids=self.gpuids)
                ),
                geo,
                angle,
                "matched",
                gpuids=self.gpuids,
            )
        )

    def run_main_iter(self):
        """
        Goes through the main iteration for the given configuration.
        :return: None
        """
        t = self.__t__
        Quameasopts = self.Quameasopts
        x_rec = copy.deepcopy(self.res)
        lambdaForTv = 2 * self.__bm__ * self.__lambda__
        for i in range(self.niter):

            res_prev = copy.deepcopy(self.res) if Quameasopts is not None else None
            if self.verbose:
                self._estimate_time_until_completion(i)

            getattr(self, self.dataminimizing)()

            x_rec_old = copy.deepcopy(x_rec)
            x_rec = im3ddenoise(self.res, self.__numiter_tv__, 1.0 / lambdaForTv, self.gpuids)
            t_old = t
            t = (self.__p__ + np.sqrt(self.__q__ + 4 * t ** 2)) / 2
            self.res = x_rec + (t_old - 1) / t * (x_rec - x_rec_old)
            
            if Quameasopts is not None:
                self.error_measurement(res_prev, i)


fista = decorator(FISTA, name="FISTA")


class ISTA(FISTA):  # noqa: D101
    __doc__ = FISTA.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        FISTA.__init__(self, proj, geo, angles, niter, **kwargs)

    def run_main_iter(self):
        """
        Goes through the main iteration for the given configuration.
        :return: None
        """
        Quameasopts = self.Quameasopts
        lambdaForTv = 2 * self.__bm__ * self.lmbda
        for i in range(self.niter):

            res_prev = copy.deepcopy(self.res) if Quameasopts is not None else None
            if self.verbose:
                self._estimate_time_until_completion(i)

            getattr(self, self.dataminimizing)()

            self.res = im3ddenoise(self.res, 20, 1.0 / lambdaForTv, self.gpuids)
            
            if Quameasopts is not None:
                self.error_measurement(res_prev, i)


ista = decorator(ISTA, name="ISTA")



from __future__ import print_function
from __future__ import with_statement
from __future__ import division

import os
import math
from re import L
import numpy
import scipy.signal
from tqdm import tqdm

from PIL import Image
from tigre.utilities.geometry import Geometry
import xml.etree.ElementTree as ET
import csv


def YXLONDataLoader(filepath, **kwargs):
    # YXLONDataLoader(filepath) Loads Nikon uCT datasets into YXLON standard
    #
    # YXLONDataLoader(filepath, OPT=VAL, ...) uses options and values.
    #    These are options in case you don't want to load the entire
    #    dataset, but only particular sets of projections.
    #    The possible arguments are:
    #         'sampling': type of sampling. default 'equidistant' Can be:
    #                'equidistant': equidistantly sample the entire set
    #                     of angles. 'num_angles' should be set for partial
    #                     loading of the data.
    #                'step': sample the entire set of projections every
    #                       'sampling_step' angles.
    #                'continous': Load the first 'num_angles' amount of
    #                           angles only.
    #
    #         'num_angles': Number of total angles to load. Default all of
    #                  them. Useful for 'equidistant' and 'continous' loading
    #
    #         'sampling_step': step to load when loading projections.
    #                 Default=1. Useful for 'step' loading.

    folder, geometry, angles = readYXLONGeometry(filepath)
    return loadYXLONProjections(folder, geometry, angles, **kwargs)


def readYXLONGeometry(filepath):

    # Developed by A. Biguri and W. Sun
    # W. Sun edited on 06.10.2018 for 3D pro version 5.2.6809.15380 (2018)
    # Modified for YXLON readers by A. Biguri

    # There are 2 types of YXLON datasets that I've seen. 
    # the ones that have ScannerParameters.xml and the ones that have a txt with the filename as scanner parameters.

    if os.path.isfile(os.path.join(filepath, "ScanParameter.xml")):
        folder, geometry, angles = __YXLON_xml_geometry(filepath)
    else:
        files = [file for file in os.listdir(filepath) if file.endswith(".txt")]
        if not files:
            raise ValueError("No Geometry file found in folder: " + folder)

        folder, geometry, angles = __YXLON_txt_geometry(filepath)
    return  folder, geometry, angles

def __YXLON_xml_geometry(filepath):

    geometry = Geometry()

    # cone beam scan parameters
    file = os.path.join(filepath, "ScanParameter.xml")
    ns = dict([node for _, node in ET.iterparse(file, events=["start-ns"])])
    tree = ET.parse(file)

    params = tree.getroot().find("ConeBeamScanParameter")
    if params is None:
        raise RuntimeError("Cone beam scan parameters not found")

    # projection information

    rotDetector = params.find("DetectorOrientation")
    if rotDetector is not None:
        # TODO: handle DetectorOrientation
        if rotDetector.text.lower() != "original":
            print("Detector orientation found but ignored!")

    flipDetector = params.find("DetectorFlipDirections")
    if flipDetector is not None:
        # TODO: handle DetectorFlipDirections
        if flipDetector.text.lower() != "none":
            print("Detector flip direction found but ignored!")

    pitch = params.find("PixelSize")
    geometry.dDetector = numpy.array(
        [float(pitch.find("Height").text), float(pitch.find("Width").text)]
    )

    size = params.find("ProjectionSize")
    geometry.nDetector = numpy.array([int(size.find("Height").text), int(size.find("Width").text)])
    geometry.sDetector = geometry.nDetector * geometry.dDetector

    # trajectory information

    trajectory = params.find("ScanTrajectory")
    if trajectory is None:
        raise RuntimeError("Unable to identify scan trajectory")

    scan_type = trajectory.attrib["{{{}}}type".format(ns["xsi"])]
    if scan_type != "CircleTrajectory":
        raise ValueError("Unsupported trajectory type: " + scan_type)

    geometry.mode = "cone"
    geometry.filter = None
    geometry.accuracy = 0.5

    scale = float(trajectory.find("PixelPitch").text)
    geometry.DSD = scale * float(trajectory.find("Fdd").text)
    geometry.DSO = scale * float(trajectory.find("Fod").text)

    # TODO: check meaning of offset
    geometry.COR = -scale * float(trajectory.find("YoCenter").text)

    geometry.offOrigin = numpy.array((0, 0, 0))

    count_proj = int(trajectory.find("NumberOfProjections").text)
    start_angle = float(trajectory.find("StartAngle").text)
    range_angle = float(trajectory.find("Rotation").text)

    angles = math.radians(start_angle) - numpy.arange(count_proj) * math.radians(
        range_angle / count_proj
    )

    dejustment = params.find("DejustmentParameter")
    if dejustment is not None:

        offsX = float(dejustment.find("HorizontalDetectorOffset").text)
        offsY = float(dejustment.find("VerticalDetectorOffset").text)
        geometry.offDetector = -scale * numpy.array((offsY, offsX))

        # TODO: check meaning of parameters
        tiltA = math.radians(float(dejustment.find("DetectorTiltA").text))
        tiltB = math.radians(float(dejustment.find("DetectorTiltB").text))
        tiltC = math.radians(float(dejustment.find("DetectorTiltC").text))
        if tiltA != 0 or tiltB != 0 or tiltC != 0:
            print("Detector tilt found but ignored! Parameter matching unknown, if you have more info email tigre.toolbox@gmail.com")

        geometry.rotDetector = numpy.array((0, 0, 0))

    else:
        geometry.offDetector = numpy.array((0, 0))
        geometry.rotDetector = numpy.array((0, 0, 0))

    geometry.nVoxel = numpy.array(
        (geometry.nDetector[0], geometry.nDetector[1], geometry.nDetector[1])
    )
    geometry.sVoxel = (
        numpy.array((geometry.sDetector[0], geometry.sDetector[1], geometry.sDetector[1]))
        * geometry.DSO
        / geometry.DSD
    )
    geometry.dVoxel = geometry.sVoxel / geometry.nVoxel
    geometry.whitelevel = float(2**params.find("ProjectionBitdepth")-1)

    return filepath, geometry, angles


def __YXLON_txt_geometry(filepath):

    files = [file for file in os.listdir(filepath) if file.endswith(".txt")]
    ini = files[0]

    geometry = Geometry()
    geometry.accuracy = 0.5
    binning=1
    with open(os.path.join(filepath, ini)) as tsv:
        for line in csv.reader(tsv, dialect="excel-tab"): #You can also use delimiter="\t" rather than giving a dialect.
            if not line:
                continue
            if line[-1]=='Number of Detector Elements ':
                geometry.nDetector=numpy.array((int(line[0]),int(line[0])))
            if line[-1]=='2D-Pixel Size [mm]':
                geometry.dDetector=numpy.array((float(line[0]),float(line[0])))
            if line[-1]=='Pixelbinning':
                binning=float(line[0])
            if line[-1]=='3D-XY-Pixel Size [mm]':
                geometry.dVoxel[1]=float(line[0])
                geometry.dVoxel[2]=float(line[0])
            if line[-1]=='3D-Z-Pixel Size [mm]':
                geometry.dVoxel[0]=float(line[0])
            if line[-1]=='Image Dimension':
                geometry.nVoxel[1]=int(line[0])
                geometry.nVoxel[2]=int(line[0])
            if line[-1]=='Number of Z-slices':
                geometry.nVoxel[0]=float(line[0])
            if line[-1]=='FOD [mm]':
                geometry.DSO=float(line[0])
            if line[-1]=='FDD [mm]':
                geometry.DSD=float(line[0])
            if line[-1]=='Center Offset [mm]':
                geometry.COR=float(line[0])
            if line[-1]=='CT: Number of Projections':
                n_angles = int(line[0])

    geometry.dDetector=geometry.dDetector * binning
    geometry.sDetector = geometry.nDetector * geometry.dDetector
    geometry.sVoxel = geometry.nVoxel * geometry.dVoxel


    ## whitelevel
    geometry.whitelevel = float(2**16-1)

    ## angles
    angles = -numpy.linspace(0,2*numpy.pi,n_angles)
    return filepath, geometry, angles


def loadYXLONProjections(folder, geometry, angles, **kwargs):
    # projections, geometry, angles = loadYXLONProjections(filepath, geometry, angles, **kwargs)
    #    loads YXLON uCT machine projections
    #
    #    loadYXLONProjections(filepath, geometry, angles) Loads a dataset given its FILEPATH,
    #       GEOMETRY and ANGLES (loaded from readYXLONGeometry())
    #       See YXLONDataLoader() for additional options.
    #
    #% developed by A. Biguri and W. Sun 06.07.2020
    # Modified by A. biguri for YXLON datasets

    # parse inputs
    angles, indices = parse_inputs(geometry, angles, **kwargs)

    # load images
    if os.path.isdir(os.path.join(folder,'Projections')):
        folder= os.path.join(folder,'Projections')
    elif os.path.isdir(os.path.join(folder,'Projektionen')):
        folder= os.path.join(folder,'Projektionen')
    else:
        raise ValueError("Folder with projections not found")

    files = sorted([file for file in os.listdir(folder) if file.lower().endswith(".raw")])
    istiff=False
    if not files:
        files = sorted([file for file in os.listdir(folder) if file.lower().endswith(".tif") or file.lower().endswith(".tiff")])
        if not files:
            raise ValueError("No projection founds with extension .raw or .tif")
        istiff=True
    
    if not istiff:
        image =  numpy.fromfile(os.path.join(folder,  files[indices[0]]), dtype=numpy.uint16)
        image = numpy.asarray(image).astype(numpy.float32)
        image = numpy.flipud(numpy.reshape(image,(geometry.nDetector)))
    else:
        image = Image.open(os.path.join(folder, files[indices[0]]))
        image = numpy.asarray(image).astype(numpy.float32)

    projections = numpy.zeros([len(indices), image.shape[0], image.shape[1]], dtype=numpy.single)
    projections[0, :, :] = -numpy.log(image / float(geometry.whitelevel))
    # Python uses zero-based indexing
    # if we start with index = 1 we get an error
    # use enumerate, it's cleaner
    print("Loading YXLON dataset: " + folder)
    for index, i in enumerate(tqdm(indices[1:]), 1):

        if not istiff:
            image =  numpy.fromfile(os.path.join(folder,  files[i]), dtype=numpy.uint16)
            image = numpy.asarray(image).astype(numpy.float32)
            image = numpy.flipud(numpy.reshape(image,(geometry.nDetector)))
        else:
            image = Image.open(os.path.join(folder, files[i]))
            image = numpy.asarray(image).astype(numpy.float32)

        projections[index, :, :] = -numpy.log(image / float(geometry.whitelevel))

    del geometry.whitelevel

    return numpy.asarray(projections), geometry, angles


def parse_inputs(geometry, angles, **kwargs):

    # TODO: warn user about invalid options or values
    sampling = kwargs["sampling"] if "sampling" in kwargs else "equidistant"
    nangles = int(kwargs["num_angles"]) if "num_angles" in kwargs else len(angles)
    step = int(kwargs["sampling_step"]) if "sampling_step" in kwargs else 1

    indices = numpy.arange(0, len(angles))

    if sampling == "equidistant":
        step = int(round(len(angles) / nangles))
        indices = indices[::step]
        angles = angles[::step]
    elif sampling == "continuous":
        indices = indices[:nangles]
        angles = angles[:nangles]
    elif sampling == "step":
        indices = indices[::step]
        angles = angles[::step]
    else:
        raise ValueError("Unknown sampling type: " + str(sampling))

    return angles, indices





from __future__ import print_function

import matplotlib
import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable


class plotProj:
    # NOTE: type help(plotImg) after importing in order to get a readable manual.
    (
        "\n"
        "plotProj(proj, dim) \n"
        "    plots figure \n"
        "default: progressive in slices following\n"
        "    axis (dim)\n"
        "Parameters \n"
        "---------- \n"
        "proj : Any 3D numpy array \n"
        "\n"
        'dim : ("U","V","T","u","v","t"), optional \n'
        '       default is "T"\n'
        "       NOTE: string arguments!"
        "\n"
        "angles: Any 1D numpy array. \n"
        "        Its length must be the same as proj.shape[0].\n"
        '        Works only when dim is "T" or "t"\n'
        "slice: int, optional\n"
        "     returns page of matrix according to index\n"
        "step: int, optional\n"
        "      Sets the step size between slice and slice."
        "      Step is 1 by default.\n"
        "savegif: string, optional\n"
        "         Saves the image as .gif with the file name\n"
        "show_plot: bool, optional\n"
        "           Sets whether to show the plot.\n"
        "           Default is None, automatically detects matplotlib backend\n"
        "           and decides whether to call plt.show.\n"
        "Examples:\n"
        "---------\n"
        "a=np.ones([3,3,3])\n"
        "plotImg(a)\n"
        ">>>returns plot along dim T\n"
        'plotImg(a,dim="v")\n'
        ">>>returns plot along dim V\n"
    )

    def __init__(
        self,
        proj,
        angles=None,
        dim=None,
        slice=None,
        step=1,
        savegif=None,
        colormap="gray",
        clims=None,
        show_plot=None,
    ):
        self.proj = proj
        self.dim = dim
        self.slice = slice
        self.dimint = None  # keeps track of what dim
        self.dimlist = ["U", "V", "T", "u", "v", "t", None]  # accepted parameters for dim
        self.step = step
        self.savegif = savegif
        self.angles = angles
        self.colormap = colormap
        if clims is None:
            self.min_val = np.amin(self.proj)
            self.max_val = np.amax(self.proj)
        else:
            self.min_val = clims[0]
            self.max_val = clims[1]
        if show_plot is None:
            # https://matplotlib.org/stable/tutorials/introductory/usage.html#backends
            backend = matplotlib.get_backend()
            if backend in [
                "GTK3Agg",
                "GTK3Cairo",
                "MacOSX",
                "nbAgg",
                "Qt4Agg",
                "Qt4Cairo",
                "Qt5Agg",
                "Qt5Cairo",
                "TkAgg",
                "TkCairo",
                "WebAgg",
                "WX",
                "WXAgg",
                "WXCairo",
                "module://ipykernel.pylab.backend_inline",
            ]:
                self.show_plot = True
            elif backend in ["agg", "cairo", "pdf", "pgf", "ps", "svg", "template"]:
                self.show_plot = False
            else:
                self.show_plot = True
        if self.step is None or self.step == 0:
            self.step = 1
        if self.savegif == "":
            self.savegif == None
        if self.slice is None:
            self.run()
        if self.slice is not None:
            self.slicer()

    def run(self):
        if self.dim not in self.dimlist and self.dim is not None:
            raise NameError("check inputs for dim, should be string.")
        if self.angles is not None and self.angles.shape[0] != self.proj.shape[0]:
            raise NameError("check inputs for angles, should be size of proj.shape[0]")

        if self.dim in ["U", "u"]:
            self.dimint = 2
            self.dimlist = ["->V", "->T", "U"]
            self.run_plot()
        if self.dim in ["V", "v"]:
            self.dimint = 1
            self.dimlist = ["->U", "->T", "V"]
            self.run_plot()
        if self.dim in [None, "T", "a"]:
            self.dimint = 0
            self.dimlist = ["->U", "->V", "T"]
            self.run_plot()

    def update_frame(self, it, fig, min_val, max_val):
        i = range(0, self.proj.shape[self.dimint])[:: self.step][it]
        fig.clf()
        axis = fig.add_subplot(1, 1, 1)
        if self.dimint == 2:
            mappable = axis.imshow(
                np.squeeze(self.proj[:, :, i]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dimint == 1:
            mappable = axis.imshow(
                np.squeeze(self.proj[:, i]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dimint == 0:
            mappable = axis.imshow(
                np.squeeze(self.proj[i]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        # axis.get_xaxis().set_ticks([])
        # axis.get_yaxis().set_ticks([])
        axis.set_xlabel(self.dimlist[0])
        axis.set_ylabel(self.dimlist[1])
        if self.angles is not None:
            axis.set_title(
                "{}:{}, alpha={:+.3f} pi".format(self.dimlist[2], i, self.angles[i] / np.pi)
            )
        else:
            axis.set_title(self.dimlist[2] + ":" + str(i))
        divider = make_axes_locatable(axis)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        fig.colorbar(mappable, cax=cax)
        # plt.pause(0.01)

    def run_plot(self):

        dim = self.proj.shape

        fig = plt.figure()
        ani = animation.FuncAnimation(
            fig,
            self.update_frame,
            fargs=(fig, self.min_val, self.max_val),
            interval=100,
            repeat_delay=1000,
            frames=len(range(0, dim[self.dimint])[:: self.step]),
        )
        if self.savegif is not None:
            ani.save(self.savegif, writer="pillow")
            self._show()
        else:
            self._show()

    def slicer(self):

        if self.dim in ["U", "u"]:
            plt.xlabel("V")
            plt.ylabel("T")
            plt.imshow(
                np.squeeze(self.proj[:, :, self.slice]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dim in ["V", "v"]:
            plt.xlabel("U")
            plt.ylabel("T")
            plt.imshow(
                np.squeeze(self.proj[:, self.slice]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        if self.dim in [None, "T", "t"]:
            if self.angles is not None:
                plt.title("alpha={:+.3f} pi".format(self.angles[self.slice] / np.pi))
            plt.xlabel("U")
            plt.ylabel("V")
            plt.imshow(
                np.squeeze(self.proj[self.slice]),
                cmap=self.colormap,
                origin="lower",
                vmin=self.min_val,
                vmax=self.max_val,
            )
        self._show()

    def _show(self):
        if self.show_plot:
            plt.show()


def plotSinogram(proj, posV, show_plot=None):  # noqa: N803
    """
    plotSinogram(proj, posV)
        plots sinogram at V=posV

    Parameters
    ----------
    proj : Any 3D numpy array
    posV : integer. in range of 0:proj.shape[1].
    """
    plotProj(proj, dim="V", slice=posV, show_plot=show_plot)


plotproj = plotProj




from tigre.utilities.cuda_interface._types cimport Geometry as c_Geometry, convert_to_c_geometry, free_c_geometry

cimport numpy as np
import numpy as np

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

from libc.stdlib cimport malloc, free
# from libc.stdio cimport printf, sprintf
# from tigre.utilities.errors import TigreCudaCallError
# cdef extern from "numpy/arrayobject.h":
#     void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)

cdef extern from "gpuUtils.hpp":
    cdef int GetGpuIdArray(char* kacGPUName, int* piDeviceIds, int iIdCountMax, char* pcMessage)
cdef extern from "gpuUtils.hpp":
    cdef void GetGpuName(int iDeviceId, char* pcName)
cdef extern from "gpuUtils.hpp":
    cdef int GetGpuCount()
# cdef extern from "gpuUtils.hpp":
#     cdef int GetMaxGpuCount()

# def cuda_raise_errors(error_code):
#     if error_code:
#         raise TigreCudaCallError('Atb:',error_code)

def getGpuIdList(nameGPU = ""):
    # print("getGpuIdList(nameGPU = {})".format(nameGPU))
    cdef int nGPUMax = GetGpuCount()
    cdef int* paIndexes    = <int*> malloc(nGPUMax* sizeof(int))
    cdef char* pcMessage    = <char*> malloc(65535* sizeof(char))
    cdef int iGPUCount = GetGpuIdArray(nameGPU.encode('utf-8'), paIndexes, nGPUMax, pcMessage)
    strMessage = (<bytes>pcMessage).decode('utf-8')
    # print("iGPUCount = {}".format(iGPUCount))
    # print(strMessage)
    if iGPUCount > 0:
        listDevices = [x for x in paIndexes[:iGPUCount]]
    else:
        listDevices = []
    free(paIndexes)
    free(pcMessage)
    return listDevices

def getGpuName(iDeviceId = 0):
    # print("iDeviceId = {}".format(iDeviceId))
    cdef char* pcDeviceName    = <char*> malloc(128* sizeof(char))
    GetGpuName(iDeviceId, pcDeviceName)
    return (<bytes>pcDeviceName).decode('utf-8')

def getGpuNames():
    nDevices = GetGpuCount()
    return [getGpuName(x) for x in range(nDevices)]

        

# deviceProp.GetMultiProcessorCount() * deviceProp.GetSMCores() * deviceProp.GetClockRate()

# multiProcessorCount




from tigre.utilities.cuda_interface._types cimport Geometry as c_Geometry, convert_to_c_geometry, free_c_geometry
from tigre.utilities.cuda_interface._gpuUtils cimport GpuIds as c_GpuIds, convert_to_c_gpuids, free_c_gpuids

cimport numpy as np
import numpy as np

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()

from libc.stdlib cimport malloc, free
from tigre.utilities.errors import TigreCudaCallError
cdef extern from "numpy/arrayobject.h":
    void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)

cdef extern from "voxel_backprojection.hpp":
    cdef int voxel_backprojection(float* projections, c_Geometry geo, float* result,float * alphas,int nalpha, const c_GpuIds& gpuids)
cdef extern from "voxel_backprojection2.hpp":
    cdef int voxel_backprojection2(float* projections, c_Geometry geo, float* result,float * alphas,int nalpha, const c_GpuIds& gpuids)
cdef extern from "voxel_backprojection_parallel.hpp":
    cdef int voxel_backprojection_parallel(float* projections, c_Geometry geo, float* result,float * alphas,int nalpha, const c_GpuIds& gpuids)


def cuda_raise_errors(error_code):
    if error_code:
        raise TigreCudaCallError('Atb:',error_code)



def _Atb_ext(np.ndarray[np.float32_t, ndim=3] projections, geometry, np.ndarray[np.float32_t, ndim=2] angles, backprojection_type="FDK", mode="cone", gpuids=None):

    cdef c_GpuIds* c_gpuids = convert_to_c_gpuids(gpuids)
    if not c_gpuids:
        raise MemoryError("Error loading gpuIds")
    
    cdef int total_projections = angles.shape[0]

    cdef c_Geometry* c_geometry = convert_to_c_geometry(geometry, total_projections)

    angles = np.ascontiguousarray(angles)

    cdef float* c_model = <float*> malloc(geometry.nVoxel[0] * geometry.nVoxel[1] * geometry.nVoxel[2] * sizeof(float))
    cdef float* c_angles = <float*> angles.data

    # TODO: Error if backprojection_type isn't FDK or matched
    if backprojection_type == "matched":
        krylov_proj = True
    elif backprojection_type == "FDK":
        krylov_proj = False
    else:
        print("Warning: Unknown backprojector, using default matched")
        krylov_proj = True

    if mode == "parallel":
        cone_beam = False
    elif mode == "cone":
        cone_beam = True
    else:
        print("Warning: Unknown mode, using default cone beam")
        cone_beam = True
        
    projections = np.ascontiguousarray(projections)

    cdef float* c_projections = <float*> projections.data

    if cone_beam:
        if krylov_proj:
            cuda_raise_errors(voxel_backprojection2(c_projections, c_geometry[0], c_model, c_angles, total_projections, c_gpuids[0]))
        else:
            cuda_raise_errors(voxel_backprojection(c_projections, c_geometry[0], c_model, c_angles, total_projections, c_gpuids[0]))


    else:
        cuda_raise_errors(voxel_backprojection_parallel(c_projections, c_geometry[0], c_model, c_angles, total_projections, c_gpuids[0]))

    cdef np.npy_intp shape[3]
    shape[0] = <np.npy_intp> geometry.nVoxel[0]
    shape[1] = <np.npy_intp> geometry.nVoxel[1]
    shape[2] = <np.npy_intp> geometry.nVoxel[2]

    # TODO: Swap axis here could be making a copy
    model = np.PyArray_SimpleNewFromData(3, shape, np.NPY_FLOAT32, c_model)
    PyArray_ENABLEFLAGS(model, np.NPY_OWNDATA) # Attribute new memory owner

    free_c_geometry(c_geometry)


    return model



import numpy as np
import tigre
from numpy.lib.stride_tricks import as_strided
from scipy.interpolate import RegularGridInterpolator


def init_multigrid(proj, geo, alpha, alg):
    # WARNING: This takes a lot of memory!
    if alg == "SART":
        italg = tigre.algorithms.sart
    if alg == "SIRT":
        italg = tigre.algorithms.sirt
    finalsize = geo.nVoxel

    maxval = max(proj.ravel())
    minval = min(proj.ravel())

    # Start with 16 (raise this for larger images)
    geo.nVoxel = np.array([16, 16, 16])
    geo.dVoxel = geo.sVoxel / geo.nVoxel
    if (geo.nVoxel > finalsize).all():
        return np.zeros(finalsize, dtype=np.float32)
    niter = 100
    initres = np.zeros(geo.nVoxel, dtype=np.float32)
    while (geo.nVoxel != finalsize).all():
        geo.dVoxel = geo.sVoxel / geo.nVoxel

        initres = italg(proj, geo, alpha, niter, init=initres, verbose=False)

        # get new dims(should be a way to do this more efficiently).

        geo.nVoxel = geo.nVoxel * 2
        geo.nVoxel[geo.nVoxel > finalsize] = finalsize[geo.nVoxel > finalsize]
        geo.dVoxel = geo.sVoxel / geo.nVoxel
        (x, y, z) = (
            np.linspace(minval, maxval, geo.nVoxel[0] / 2, dtype=np.float32),
            np.linspace(minval, maxval, geo.nVoxel[1] / 2, dtype=np.float32),
            np.linspace(minval, maxval, geo.nVoxel[2] / 2, dtype=np.float32),
        )

        # evaluate the function sart at the points xv,yv,zv

        xv, yv, zv = [
            tile_array(tile_array(x, 2), geo.nVoxel[0] ** 2),
            tile_array(tile_array(y, 2), geo.nVoxel[0] ** 2),
            tile_array(tile_array(x, 2), geo.nVoxel[0] ** 2),
        ]

        initres = RegularGridInterpolator((x, y, z), initres)(np.column_stack((xv, yv, zv)))
        initres = initres.reshape(geo.nVoxel)

    return initres


def tile_array(mat, b1):
    (r,) = mat.shape
    (rs,) = mat.strides
    x = as_strided(mat, (r, b1), (rs, 0))
    return x.reshape(r * b1)



import matplotlib.patches
import numpy as np
import tigre
from mpl_toolkits.mplot3d import art3d

# https://matplotlib.org/mpl_toolkits/mplot3d/tutorial.html


class Arrow3D(matplotlib.patches.FancyArrowPatch):
    def __init__(self, xs, ys, zs, *args, **kwargs):
        matplotlib.patches.FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        from mpl_toolkits.mplot3d import proj3d

        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        matplotlib.patches.FancyArrowPatch.draw(self, renderer)

    def do_3d_projection(self, render=None):
        from mpl_toolkits.mplot3d import proj3d
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        return np.min(zs)


ROT_DEFAULT = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
TRANS_DEFAULT = np.array([0, 0, 0])


def pathpatch_2d_to_3d_affine(pathpatch, mat_rot=ROT_DEFAULT, vec_trans=TRANS_DEFAULT):
    """
    Transforms a 2D Patch to a 3D patch using the affine tranform
    of the given rotation matrix and translation vector.
    The pathpatch is assumed to be on the plane Z = 0.
    """
    path = pathpatch.get_path()  # Get the path and the associated transform
    trans = pathpatch.get_patch_transform()

    path = trans.transform_path(path)  # Apply the transform

    pathpatch.__class__ = art3d.PathPatch3D  # Change the class
    pathpatch._code3d = path.codes  # Copy the codes
    pathpatch._facecolor3d = pathpatch.get_facecolor  # Get the face color

    verts = path.vertices  # Get the vertices in 2D

    M = np.array(
        [
            [mat_rot[0, 0], mat_rot[0, 1], mat_rot[0, 2], vec_trans[0]],
            [mat_rot[1, 0], mat_rot[1, 1], mat_rot[1, 2], vec_trans[1]],
            [mat_rot[2, 0], mat_rot[2, 1], mat_rot[2, 2], vec_trans[2]],
        ]
    )

    pathpatch._segment3d = np.array([np.dot(M, (x, y, 0, 1)) for x, y in verts])


def plot_geometry(geo, angle=0):
    """Plots the given geometry."""
    import mpl_toolkits.mplot3d.art3d as art3d
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=(6, 6))
    fig.suptitle("Cone Beam Compute Tomography geometry")
    ax = fig.add_subplot(111, projection="3d")
    ax.set_title("Current CBCT geometry, in scale")

    limXY = max(geo.DSO, geo.DSD - geo.DSO)
    limZ = geo.sVoxel[0]

    ax.set_box_aspect((limXY, limXY, limZ))

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    ax.set_xlim3d(-limXY * 1.2, limXY * 1.2)
    ax.set_ylim3d(-limXY * 1.2, limXY * 1.2)
    ax.set_zlim3d(-limZ * 1.2, limZ * 1.2)

    # Trajectory of Source
    # https://matplotlib.org/devdocs/api/_as_gen/matplotlib.patches.Circle.htm
    circ = matplotlib.patches.Circle((0, 0), geo.DSO, color="black", fill=False, ls="-.", lw=0.5)
    ax.add_patch(circ)
    art3d.pathpatch_2d_to_3d(circ, z=0, zdir="z")

    # Trajectory of Detector
    circ = matplotlib.patches.Circle(
        (0, 0), geo.DSD - geo.DSO, color="black", fill=False, ls="-.", lw=0.5
    )
    ax.add_patch(circ)
    art3d.pathpatch_2d_to_3d(circ, z=0, zdir="z")

    # Source
    # ax.scatter([0], [0], [0], color="g", s=100)
    sourcePos3D = [geo.DSO * np.cos(angle), geo.DSO * np.sin(angle), 0]  # xyz
    ax.scatter([sourcePos3D[0]], [sourcePos3D[1]], [sourcePos3D[2]], color="steelblue", s=100)

    # Axes XYZ
    length_axis = geo.sVoxel[0]
    x_axis = Arrow3D(
        [0, length_axis],
        [0, 0],
        [0, 0],
        mutation_scale=10,
        shrinkA=0,
        lw=1,
        arrowstyle="-|>",
        color="r",
    )
    y_axis = Arrow3D(
        [0, 0],
        [0, length_axis],
        [0, 0],
        mutation_scale=10,
        shrinkA=0,
        lw=1,
        arrowstyle="-|>",
        color="b",
    )
    z_axis = Arrow3D(
        [0, 0],
        [0, 0],
        [0, length_axis],
        mutation_scale=10,
        shrinkA=0,
        lw=1,
        arrowstyle="-|>",
        color="g",
    )
    ax.add_artist(x_axis)
    ax.add_artist(y_axis)
    ax.add_artist(z_axis)

    # Detector
    print("sDetector = {}".format(geo.sDetector))
    alpha_detector = 0.7
    detectorPos3D = [
        (geo.DSD - geo.DSO) * np.cos(angle + np.pi) + geo.offDetector[1] * np.sin(angle + np.pi),
        (geo.DSD - geo.DSO) * np.sin(angle + np.pi) - geo.offDetector[1] * np.cos(angle + np.pi),
        geo.offDetector[0],
    ]  # xyz
    detector = matplotlib.patches.Rectangle(
        (-geo.sDetector[1] / 2, -geo.sDetector[0] / 2),
        geo.sDetector[1],
        geo.sDetector[0],
        angle=0,  # angle,
        color="maroon",
        fill=True,
        alpha=alpha_detector,
    )
    print("detector={}".format(detector))
    ax.add_patch(detector)
    mat_rot = np.array(
        [[-np.sin(angle), 0, np.cos(angle)], [np.cos(angle), 0, np.sin(angle)], [0, 1, 0]]
    )
    pathpatch_2d_to_3d_affine(detector, mat_rot, detectorPos3D)

    # Image FOV
    alpha_img = 0.1
    offOrigin = np.array([geo.offOrigin[2], geo.offOrigin[1], geo.offOrigin[0]])
    mat_rot_xy = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float)
    for idx in range(2):
        img_face_xy = matplotlib.patches.Rectangle(
            (-geo.sVoxel[2] / 2, -geo.sVoxel[1] / 2),  # xy order
            geo.sVoxel[2],
            geo.sVoxel[1],
            angle=0,  # angle,
            color="black",
            fill=True,
            alpha=alpha_img,
        )
        ax.add_patch(img_face_xy)
        pathpatch_2d_to_3d_affine(
            img_face_xy,
            mat_rot_xy,
            np.array([0, 0, -geo.sVoxel[0] / 2 if idx == 0 else geo.sVoxel[0] / 2]) + offOrigin,
        )

    mat_rot_yz = np.array([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=np.float)
    for idx in range(2):
        img_face_yz = matplotlib.patches.Rectangle(
            (-geo.sVoxel[1] / 2, -geo.sVoxel[0] / 2),  # xy order
            geo.sVoxel[1],
            geo.sVoxel[0],
            angle=0,  # angle,
            color="black",
            fill=True,
            alpha=alpha_img,
        )
        ax.add_patch(img_face_yz)
        pathpatch_2d_to_3d_affine(
            img_face_yz,
            mat_rot_yz,
            np.array([-geo.sVoxel[2] / 2 if idx == 0 else geo.sVoxel[2] / 2, 0, 0]) + offOrigin,
        )

    mat_rot_zx = np.array([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=np.float)
    for idx in range(2):
        img_face_zx = matplotlib.patches.Rectangle(
            (-geo.sVoxel[0] / 2, -geo.sVoxel[2] / 2),  # xy order
            geo.sVoxel[0],
            geo.sVoxel[2],
            angle=0,  # angle,
            color="black",
            fill=True,
            alpha=alpha_img,
        )
        ax.add_patch(img_face_zx)
        pathpatch_2d_to_3d_affine(
            img_face_zx,
            mat_rot_zx,
            np.array([0, -geo.sVoxel[1] / 2 if idx == 0 else geo.sVoxel[1] / 2, 0]) + offOrigin,
        )

    plt.show()


if __name__ == "__main__":
    n_voxel_z = 128
    n_voxel_y = 256
    n_voxel_x = 512
    n_detector_u = 400
    n_detector_v = 300
    off_detector_u = 0  # 500
    off_detector_v = 0  # 500
    off_origin_x = 0  # 300
    off_origin_y = 0  # 100
    off_origin_z = 0  # 100

    geo = tigre.geometry(mode="cone", default=True)
    geo.nVoxel = np.array([n_voxel_z, n_voxel_y, n_voxel_x])
    geo.sVoxel = geo.nVoxel
    geo.dVoxel = geo.sVoxel / geo.nVoxel
    geo.nDetector = np.array([n_detector_v, n_detector_u])
    geo.sDetector = geo.nDetector * geo.dDetector
    geo.offDetector = np.array([off_detector_v, off_detector_u])
    geo.offOrigin = np.array([off_origin_z, off_origin_y, off_origin_x])
    print(geo)
    angle = -np.pi / 6
    plot_geometry(geo, angle)



from __future__ import print_function
from __future__ import with_statement

import os
import math
import numpy
from tqdm import tqdm

from PIL import Image
from configparser import ConfigParser, Error
from tigre.utilities.geometry import Geometry


def BrukerDataLoader(filepath, **kwargs):
    # BrukerDataLoader(filepath) Loads Bruker Skyscan datasets into TIGRE standard
    #
    # BrukerDataLoader(filepath, OPT=VAL, ...) uses options and values.
    #    These are options in case you don't want to load the entire
    #    dataset, but only particular sets of projections.
    #    The possible arguments are:
    #         'dataset_number': Some fodlers will have several scans.
    #                 Set to 'all' to load all of them, or give a number (starting from 0)
    #         'sampling': type of sampling. default 'equidistant' Can be:
    #                'equidistant': equidistantly sample the entire set
    #                     of angles. 'num_angles' should be set for partial
    #                     loading of the data.
    #                'step': sample the entire set of projections every
    #                       'sampling_step' angles.
    #                'continous': Load the first 'num_angles' amount of
    #                           angles only.
    #
    #         'num_angles': Number of total angles to load. Default all of
    #                  them. Useful for 'equidistant' and 'continous' loading
    #
    #         'sampling_step': step to load when loading projections.
    #                 Default=1. Useful for 'step' loading.

    dataset_number = kwargs["dataset_number"] if "dataset_number" in kwargs else None
    if dataset_number == "all":
        print("Loading all scans in folder, assuming the same geometry for all \n\n")
        del kwargs["dataset_number"]
        num_scans = find_number_of_datasets(filepath)
        angles = [None] * num_scans
        projections = [None] * num_scans
        for scan in range(num_scans):
            print("Loading scan number " + str(scan) + "...\n")
            folder, geometry, angles[scan] = read_Bruker_geometry(
                filepath, dataset_number=scan, **kwargs
            )
            projections[scan], geometry, angles[scan] = load_Bruker_projections(
                folder, geometry, angles[scan], dataset_number=scan, **kwargs
            )
        projections = numpy.concatenate(projections)
        angles = numpy.concatenate(angles)

        return projections, geometry, angles

    else:

        folder, geometry, angles = read_Bruker_geometry(filepath, **kwargs)
        return load_Bruker_projections(folder, geometry, angles, **kwargs)


def read_Bruker_geometry(filepath, **kwargs):

    # check if input was log file itself, or just the folder
    if filepath.endswith(".log"):
        folder, ini = os.path.split(filepath)
    else:
        folder = filepath
        files = [file for file in os.listdir(folder) if file.endswith(".log")]
        if not files:
            raise ValueError("No .log file found in folder: " + folder)

        num_scans = find_number_of_datasets(filepath)

        if num_scans is not None:
            dataset_number = kwargs["dataset_number"] if "dataset_number" in kwargs else None
            if dataset_number is None:
                raise ValueError(
                    "This folder contains many datasets, please select which one to load with BrukerDataLoader(..., dataset_number=a_number)"
                )
            if dataset_number >= num_scans:
                raise ValueError("Dataset number given larger than total number of datasets")

            matching = [s for s in files if "{0:0=2d}".format(dataset_number) + ".log" in s]
            if len(matching) > 1:
                raise AssertionError(
                    "More than 1 file for the same dataset found, confused what to do, so I error"
                )
            ini = matching[0]
        else:
            ini = files[0]
    # create configureation parser
    cfg = ConfigParser()
    cfg.read(os.path.join(folder, ini))
    cfg_system = cfg["System"]

    # start empty geometry
    geometry = Geometry()
    geometry.accuracy = 0.5

    ## Detector information
    # Size of pixels in the detector

    try:
        ratio = float(cfg_system["CameraXYRatio"])
    except KeyError:
        ratio = float(cfg_system["Camera X/Y Ratio"])
    except:
        ratio = 1
    # need to verify if this is correct.
    ratio = 1

    cfg_aq = cfg["Acquisition"]
    try:
        binu = float(cfg_aq["Camera binning"][0])
        binv = float(cfg_aq["Camera binning"][2])
    except:
        binu = 1
        binv = 1

    geometry.dDetector = numpy.array(
        (
            float(cfg_system["Camera Pixel Size (um)"]) / 1000.0 * binv * ratio,
            float(cfg_system["Camera Pixel Size (um)"]) / 1000.0 * binu,
        )
    )
    # Number of pixel in the detector
    geometry.nDetector = numpy.array(
        (float(cfg_aq["Number of Rows"]), float(cfg_aq["Number of Columns"]))
    )

    # Total size of the detector
    geometry.sDetector = geometry.nDetector * geometry.dDetector
    geometry.nDetector = geometry.nDetector.astype(int)
    ## Offset of the detector:
    try:
        offs = (
            -(geometry.nDetector[0] / 2 - float(cfg_aq["Optical Axis (line)"]))
            * geometry.dDetector[0]
        )
    except:
        offs = 0.0

    geometry.offDetector = numpy.array((offs, 0.0))

    # Size of each voxel
    geometry.dVoxel = numpy.array(
        (
            float(cfg_aq["Image Pixel Size (um)"]) / 1000,
            float(cfg_aq["Image Pixel Size (um)"]) / 1000,
            float(cfg_aq["Image Pixel Size (um)"]) / 1000,
        )
    )
    geometry.nVoxel = numpy.array(
        (geometry.nDetector[0], geometry.nDetector[1], geometry.nDetector[1])
    )
    geometry.sVoxel = geometry.nVoxel * geometry.dVoxel
    geometry.nVoxel = geometry.nVoxel.astype(int)

    #% Global geometry
    geometry.DSO = float(cfg_aq["Object to Source (mm)"])
    geometry.DSD = float(cfg_aq["Camera to Source (mm)"])

    geometry.whitelevel = 2 ** int(cfg_aq["Depth (bits)"])

    angles = numpy.arange(
        0.0,
        float(cfg_aq["Number of Files"]) * float(cfg_aq["Rotation Step (deg)"]),
        float(cfg_aq["Rotation Step (deg)"]),
    )
    angles = angles[:-1] * numpy.pi / 180

    files = [file for file in os.listdir(folder) if file.endswith(".csv")]
    if files:
        offset = numpy.genfromtxt(
            os.path.join(folder, files[0]), delimiter=",", skip_header=5, dtype=float
        )
        offset = numpy.delete(offset, 0, 1)
        geometry.offDetector = geometry.offDetector + numpy.fliplr(offset) * (
            geometry.dDetector / [binv, binu]
        )

    return filepath, geometry, angles


def load_Bruker_projections(folder, geometry, angles, **kwargs):

    angles, indices, dataset_number = parse_inputs(geometry, angles, **kwargs)

    # load images
    files = sorted([file for file in os.listdir(folder) if file.lower().endswith(".tif")])
    if dataset_number is not None:
        files = [file for file in files if file[-10:-8] == "{0:0=2d}".format(dataset_number)]

    image = Image.open(os.path.join(folder, files[indices[0]]))
    image = numpy.asarray(image).astype(numpy.float32)
    projections = numpy.zeros([len(indices), image.shape[0], image.shape[1]], dtype=numpy.single)
    projections[0, :, :] = -numpy.log(image / float(geometry.whitelevel))
    index = 1

    print("Loading Bruker Skyscan dataset: " + folder)
    for i in tqdm(indices[1:]):
        image = Image.open(os.path.join(folder, files[i]))
        image = numpy.asarray(image).astype(numpy.float32)
        projections[index, :, :] = -numpy.log(image / float(geometry.whitelevel))
        index = index + 1
    del geometry.whitelevel

    return numpy.asarray(projections), geometry, angles


## This should be on a separate "io_common.py" file.
def parse_inputs(geometry, angles, **kwargs):

    # TODO: warn user about invalid options or values
    sampling = kwargs["sampling"] if "sampling" in kwargs else "equidistant"
    nangles = int(kwargs["num_angles"]) if "num_angles" in kwargs else len(angles)
    step = int(kwargs["sampling_step"]) if "sampling_step" in kwargs else 1
    dataset_number = kwargs["dataset_number"] if "dataset_number" in kwargs else None

    indices = numpy.arange(0, len(angles))

    if sampling == "equidistant":
        step = int(round(len(angles) / nangles))
        indices = indices[::step]
        angles = angles[::step]
    elif sampling == "continuous":
        indices = indices[:nangles]
        angles = angles[:nangles]
    elif sampling == "step":
        indices = indices[::step]
        angles = angles[::step]
    else:
        raise ValueError("Unknown sampling type: " + str(sampling))

    return angles, indices, dataset_number


def find_number_of_datasets(filepath):
    # check if input was log file itself, or just the folder
    if filepath.endswith(".log"):
        folder, ini = os.path.split(filepath)
    else:
        folder = filepath
        files = [file for file in os.listdir(folder) if file.endswith(".log")]
        if not files:
            raise ValueError("No .log file found in folder: " + folder)

        ini = min(files, key=len)  # shortest one is the main one?
        cfg = ConfigParser()
        cfg.read(os.path.join(folder, ini))
        cfg_aq = cfg["Acquisition"]

        if cfg.has_option("Acquisition", "Number of connected scans"):
            return int(cfg_aq["Number of connected scans"])
        else:
            return None



from __future__ import division, absolute_import, print_function

# fix for DLL import issue on python 3.8 or higher on windows
# related to: 
# https://github.com/CERN/TIGRE/issues/247
# https://github.com/CERN/TIGRE/issues/349
import os

if hasattr(os, "add_dll_directory"):
    # Add all the DLL directories manually
    # see:
    # https://docs.python.org/3.8/whatsnew/3.8.html#bpo-36085-whatsnew
    # https://stackoverflow.com/a/60803169/19344391
    dll_directory = os.path.dirname(__file__)
    os.add_dll_directory(dll_directory)

    # The user must install the CUDA Toolkit
    cuda_bin = os.path.join(os.environ["CUDA_PATH"], "bin")
    os.add_dll_directory(cuda_bin)

from .utilities.geometry import geometry
from .utilities.geometry_default import ConeGeometryDefault as geometry_default
from .utilities.Ax import Ax
from .utilities.Atb import Atb
from .utilities.visualization.plotproj import plotproj, plotProj, plotSinogram
from .utilities.visualization.plotimg import plotimg, plotImg
from .utilities.visualization.plot_geometry import plot_geometry
from .utilities.visualization.plot_angles import plot_angles
from .utilities.CTnoise import add
from .utilities.common_geometry import staticDetectorGeo, staticDetLinearSourceGeo
from . import algorithms



from __future__ import division
from __future__ import print_function

import inspect

import numpy as np
import numpy.matlib as matlib
import tigre


class Geometry(object):
    def __init__(self):
        self.mode = None
        self.n_proj = None
        self.angles = None
        self.filter = None
        self.nVoxel = np.zeros(3)
        self.dVoxel = np.zeros(3)
        self.sVoxel = np.zeros(3)
        self.nDetector = np.zeros(2)
        self.dDetector = np.zeros(2)
        self.sDetector = np.zeros(2)


    def check_geo(self, angles, verbose=False):
        if angles.ndim == 1:
            self.n_proj = angles.shape[0]
            zeros_array = np.zeros((self.n_proj, 1), dtype=np.float32)
            self.angles = np.hstack((angles.reshape(self.n_proj, 1), zeros_array, zeros_array))

        elif angles.ndim == 2:
            if angles.shape[1] != 3:
                raise BufferError("Expected angles of dimensions (n, 3), got: " + str(angles.shape))
            self.n_proj = angles.shape[0]
            self.angles = angles.copy()
        else:
            raise BufferError("Unexpected angles shape: " + str(angles.shape))
        if self.mode is None:
            self.mode = "cone"

        mandatory_attribs = [
            "nVoxel",
            "sVoxel",
            "dVoxel",
            "nDetector",
            "sDetector",
            "dDetector",
            "DSO",
            "DSD",
        ]
        included_attribs_indx = [hasattr(self, attrib) for attrib in mandatory_attribs]
        if not all(included_attribs_indx):
            raise AttributeError(
                "following manditory fields "
                "missing from geometry:"
                + str([attrib for attrib in mandatory_attribs if not hasattr(self, attrib)])
            )
        optional_attribs = ["offOrigin", "offDetector", "rotDetector", "COR", "mode", "accuracy"]

        # image data
        if not self.nVoxel.shape == (3,):
            raise AttributeError("geo.nVoxel.shape should be (3, )")
        if not self.sVoxel.shape == (3,):
            raise AttributeError("geo.sVoxel.shape should be (3, )")
        if not self.dVoxel.shape == (3,):
            raise AttributeError("geo.dVoxel.shape should be (3, )")
        if not sum(abs(self.dVoxel * self.nVoxel - self.sVoxel)) < 1e-6:
            "nVoxel*dVoxel is not equal to sVoxel. " "Check fields."

        # Detector Data
        if not self.nDetector.shape == (2,):
            raise AttributeError("geo.nDecetor.shape should be (2, )")
        if not self.sDetector.shape == (2,):
            raise AttributeError("geo.sDetector.shape should be (2, )")
        if not self.dDetector.shape == (2,):
            raise AttributeError("geo.dDetector.shape should be (2, )")
        if not sum(abs(self.dDetector * self.nDetector - self.sDetector)) < 1e-6:
            raise AttributeError("nDetector*dDetecor is not equal to sDetector. Check fields.")

        for attrib in ["DSD", "DSO"]:
            self.__check_and_repmat__(attrib, angles)

        if hasattr(self, "offOrigin"):
            self.__check_and_repmat__("offOrigin", angles)
        else:
            self.offOrigin = np.array([0, 0, 0])
            self.__check_and_repmat__("offOrigin", angles)

        if hasattr(self, "offDetector"):
            self.__check_and_repmat__("offDetector", angles)
        else:
            self.offDetector = np.array([0, 0])
            self.__check_and_repmat__("offDetector", angles)

        if hasattr(self, "rotDetector"):
            self.__check_and_repmat__("rotDetector", angles)
        else:
            self.rotDetector = np.array([0, 0, 0])
            self.__check_and_repmat__("rotDetector", angles)

        if hasattr(self, "COR"):
            self.__check_and_repmat__("COR", angles)
        else:
            self.COR = np.zeros(angles.shape[0])

        if not hasattr(self, "accuracy"):
            self.accuracy = 0.5
        # IMPORTANT: cast all numbers to float32
        if verbose:
            self._verbose_output()

    def checknans(self):
        for attrib in self.__dict__:
            if str(getattr(self, attrib)) == "nan":
                raise ValueError("nan found for Geometry abbtribute:" + attrib)
            elif type(getattr(self, attrib)) == np.ndarray:
                if np.isnan(getattr(self, attrib)).all():
                    raise ValueError("Nan found in Geometry abbtribute:" + attrib)

    def cast_to_single(self):
        """
        Casts all number values in current instance to single prevision floating point types.

        :return: None
        """
        for attrib in self.__dict__:
            if getattr(self, attrib) is not None:
                try:
                    setattr(self, attrib, np.float32(getattr(self, attrib)))
                except ValueError:
                    pass

    def __check_and_repmat__(self, attrib, angles):
        """
        Checks whether the attribute is a single value and repeats it into an array if it is

        :rtype: None
        :param attrib: string
        :param angles: np.ndarray
        """
        old_attrib = getattr(self, attrib)

        if type(old_attrib) in [float, int, np.float32, np.float64, np.int32]:
            new_attrib = matlib.repmat(old_attrib, 1, angles.shape[0])[0]
            setattr(self, attrib, new_attrib)

        elif type(old_attrib) == np.ndarray:
            if old_attrib.ndim == 1:
                if old_attrib.shape in [(3,), (2,)] and attrib not in ["DSD", "DSO", "COR"]:
                    new_attrib = matlib.repmat(old_attrib, angles.shape[0], 1)
                    setattr(self, attrib, new_attrib)
                elif old_attrib.shape in [(1,)]:
                    new_attrib = matlib.repmat(old_attrib, angles.shape[0], 1)
                    setattr(self, attrib, new_attrib)
                elif old_attrib.shape == (angles.shape[0],):
                    pass
            else:
                if old_attrib.shape == (angles.shape[0], old_attrib.shape[1]):
                    pass
                else:
                    raise AttributeError(
                        attrib
                        + " with shape: "
                        + str(old_attrib.shape)
                        + " not compatible with shapes: "
                        + str(
                            [
                                (angles.shape[0],),
                                (angles.shape[0], old_attrib.shape[1]),
                                (3,),
                                (2,),
                                (1,),
                            ]
                        )
                    )

        else:
            raise TypeError(
                "Data type not understood for: geo."
                + attrib
                + " with type = "
                + str(type(getattr(self, attrib)))
            )

    def _verbose_output(self):
        for obj in inspect.getmembers(self):
            if obj[0][0] == "_":
                pass
            elif obj[0] == "check_geo":
                pass
            elif type(obj[1]) == np.ndarray:
                print(self.mode + ": " + str((obj[0], obj[1].shape)))
            else:
                print(self.mode + ": " + str(obj))

    def __str__(self):
        parameters = []
        parameters.append("TIGRE parameters")
        parameters.append("-----")
        parameters.append("Geometry parameters")
        parameters.append("Distance from source to detector (DSD) = " + str(self.DSD) + " mm")
        parameters.append("Distance from source to origin (DSO)= " + str(self.DSO) + " mm")
        parameters.append("-----")
        parameters.append("Detector parameters")
        parameters.append("Number of pixels (nDetector) = " + str(self.nDetector))
        parameters.append("Size of each pixel (dDetector) = " + str(self.dDetector) + " mm")
        parameters.append("Total size of the detector (sDetector) = " + str(self.sDetector) + " mm")
        parameters.append("-----")
        parameters.append("Image parameters")
        parameters.append("Number of voxels (nVoxel) = " + str(self.nVoxel))
        parameters.append("Total size of the image (sVoxel) = " + str(self.sVoxel) + " mm")
        parameters.append("Size of each voxel (dVoxel) = " + str(self.dVoxel) + " mm")

        parameters.append("-----")
        if hasattr(self, "offOrigin") and hasattr(self, "offDetector"):
            parameters.append("Offset correction parameters")
            if hasattr(self, "offOrigin"):
                parameters.append(
                    "Offset of image from origin (offOrigin) = " + str(self.offOrigin) + " mm"
                )
            if hasattr(self, "offDetector"):
                parameters.append(
                    "Offset of detector (offDetector) = " + str(self.offDetector) + " mm"
                )

        parameters.append("-----")
        parameters.append("Auxillary parameters")
        parameters.append(
            "Samples per pixel of forward projection (accuracy) = " + str(self.accuracy)
        )

        if hasattr(self, "rotDetector"):
            parameters.append("-----")
            parameters.append(
                "Rotation of the Detector (rotDetector) = " + str(self.rotDetector) + " rad"
            )

        if hasattr(self, "COR"):
            parameters.append("-----")
            parameters.append("Centre of rotation correction (COR) = " + str(self.COR) + " mm")

        return "\n".join(parameters)

    def __cmp__(self, other):
        resultofnumpiesanallyretentiveattemptatbeingphilosophical = []
        for attrib in self.__dict__:
            result = getattr(self, attrib) == getattr(other, attrib)
            try:
                resultofnumpiesanallyretentiveattemptatbeingphilosophical.append(result.all())
            except Exception:
                try:
                    resultofnumpiesanallyretentiveattemptatbeingphilosophical.extend(result)
                except Exception:
                    resultofnumpiesanallyretentiveattemptatbeingphilosophical.append(result)

        # why is this boolean reversed when returned?
        # because for some reason its reversed when i return it from this function. Who knows.

        return not all(resultofnumpiesanallyretentiveattemptatbeingphilosophical)


class ParallelGeo(Geometry):
    def __init__(self, nVoxel):  # noqa: N803
        if nVoxel is None:
            raise ValueError("nVoxel needs to be given for initialisation of parallel beam")
        Geometry.__init__(self)
        self.mode = "parallel"

        self.nVoxel = nVoxel
        self.dVoxel = np.array([1, 1, 1])
        self.sVoxel = self.nVoxel

        self.DSO = np.float32(self.nVoxel[0])
        self.DSD = np.float32(self.nVoxel[0] * 2)

        self.dDetector = np.array([1, 1])
        self.nDetector = np.zeros((2,), dtype=np.int64)
        self.nDetector[0] = self.nVoxel[0]
        self.nDetector[1] = max(self.nVoxel[1], self.nVoxel[2])
        # self.nDetector = self.nVoxel[:2]
        self.sDetector = self.nDetector

        self.accuracy = 0.5

        self.offOrigin = np.array([0, 0, 0])
        self.offDetector = np.array([0, 0])

        self.rotDetector = np.array([0, 0, 0])


def geometry(mode="cone", nVoxel=None, default=False, high_resolution=True):
    """
    Constructor for geometry used in reconstruction of images in TIGRE

    Parameters
    ----------
    :param mode: (str)
        'cone' or 'parallel'
    :param nVoxel: (np.ndarray)
        number of voxels the reconstruction is composed of
    :param default: (bool)
        calculates other parameters in geometry. is by default true for
        parallel geometry
    :param high_resolution: (bool)
        preset values for geometry in mode=cone. WARNING: for smaller
        tests it is better to use this rather than setting nVoxel
        manually.
        if true: nVoxel = np.array([512,512,512])
        if false: nVoxel = np.array([64,64,64])
    :return: (tigre.geometry.Geometry)

    Usage
    -----
    >>> import tigre
    >>> #Cone beam with no preset parameters
    >>> geo_cone = tigre.geometry(mode='cone')
    >>> # Cone beam default, low quality
    >>> geo_cone_default = tigre.geometry(mode='cone',high_resolution=False)
    >>> # Cone beam with specific nVoxel requirements
    >>> geo_cone__default2 = tigre.geometry(nVoxel=np.array([64,64,64]),
    >>>                                     mode='cone'
    >>>                                     default=True)
    >>> # Parallel beam
    >>> geo_par = tigre.geometry(mode='parallel',
    >>>                          nVoxel=np.array([64,64,64]))
    """
    if mode == "cone":
        if default:
            return tigre.geometry_default(high_resolution, nVoxel)
        else:
            return Geometry()
    if mode == "parallel":
        return ParallelGeo(nVoxel)
    else:
        raise ValueError("mode: " + mode + " not recognised.")



import numpy as np

# Tomosymthesis
def staticDetectorGeo(geo,angles,rot=0):
    # angles: angle off the aixs between the source and detector centre (on x-y plane)
    #         when angles=0, the source is perpendicular to the detector
    # rot: rotation of both source and detector around origin
    R = geo.DSD-geo.DSO
    rot = rot/180*np.pi
    geo.DSD = geo.DSO + R*np.cos(angles)
    geo.offDetector = np.vstack([0*angles, R*np.sin(angles)]).T
    geo.rotDetector = np.vstack([0*angles, 0*angles, -angles]).T
    angles += rot
    geo.angles = np.vstack([angles, 0*angles, 0*angles]).T
    
    return geo

# Linear scan of source
def staticDetLinearSourceGeo(geo,s_pos,s_rot=0,rot=0):
    # s_pos: distance along source scanning linear trajectry 
    #        when s_pos = 0, source is aligned to the origin and detector centre
    # s_rot: rotation angle between the source linear trajectry and detector on (x-y) plane 
    # rot: source and detector rotation angle around the origin (on x-y plane)
    if np.isscalar(s_rot):
        s_rot = s_rot * np.ones_like(s_pos)
    s_rot = s_rot / 180 *np.pi
    if np.isscalar(rot):
        rot = rot * np.ones_like(s_pos)
    rot = rot/180*np.pi 
               
    ang  = np.arctan2(s_pos*np.cos(s_rot), geo.DSO + s_pos*np.sin(s_rot))
    R = geo.DSD - geo.DSO
    geo.offDetector = np.array([ang*0, R*np.sin(ang)]).T
    geo.rotDetector = np.array([0*ang, 0*ang, -ang]).T    
    geo.DSO = np.sqrt((s_pos*np.cos(s_rot))**2 + (geo.DSO + s_pos*np.sin(s_rot))**2)
    geo.DSD = geo.DSO + R*np.cos(ang)
    ang += rot        
    geo.angles = np.vstack([ang, 0*ang, 0*ang]).T
    return geo, ang



"""
1. Check the names of installed GPUs.
   names = getGpuNames()
   print(names)
2. Select the name of the GPUs that you want your program run on.
   gpuids = getGpuIds('GeForce RTX 2080 Ti')
3. Pass the gpuids instance to Tigre functions.
"""
import _gpuUtils as gpuUtils


class GpuIds(object):
    """
    A class that holds the IDs and their (common) name
    ```
    gpuids = GpuIds('GeForce RTX 2080 Ti')
    ```
    If name is ""/None, IDs of all GPUs are hold.
    ```
    gpuids = GpuIds()
    ```
    """

    def __init__(self, nameGPU=None):  # noqa: N803
        if nameGPU is None:
            self.name = ""
        else:
            self.name = nameGPU
        self.devices = gpuUtils.getGpuIdList(self.name)

    def __len__(self):
        return len(self.devices)

    def __str__(self):
        dictTemp = {
            "name": self.name,
            "devices": self.devices,
        }
        return dictTemp.__str__()


def getGpuNames():
    """Returns a list of all installed GPUs."""
    return gpuUtils.getGpuNames()


# def getGpuCount():
#     """Returns the number of installed GPUs."""
#     return gpuUtils.getGpuCount()


def getGpuIds(gpuName=None):  # noqa: N803
    """Returns the GpuIds object, which contains the IDs of devices whose name matches gpuName."""
    gpuids = GpuIds(gpuName)
    return gpuids



import copy

from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator
from tigre.utilities.im_3d_denoise import im3ddenoise



class SART(IterativeReconAlg):  
    __doc__ = (
        "SART solves Cone Beam CT image reconstruction using \n"
        "Simultaneous Algebraic Reconstruction Technique algorithm\n"
        "SART(PROJ,GEO,ALPHA,NITER) solves the reconstruction problem\n"
        "using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "to the geometry described in GEO, using NITER iterations. \n"
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        if "blocksize" in kwargs and kwargs['blocksize']>1:
            print('Warning: blocksize is set to 1, please use an OS version of the algorithm for blocksize > 1')
        kwargs.update(dict(blocksize=1))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)


sart = decorator(SART, name="sart")


class SIRT(IterativeReconAlg):  
    __doc__ = (
        "SIRT solves Cone Beam CT image reconstruction using \n"
        "Simultaneous Iterrative Reconxtructive Technique algorithm\n"
        "SIRT(PROJ,GEO,ALPHA,NITER) solves the reconstruction problem\n"
        "using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "to the geometry descrived in GEO, using NITER iterations.\n"
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        if "blocksize" in kwargs and kwargs['blocksize']>1:
            print('Warning: blocksize is set to {}, please do not specify blocksize for this algorithm'.format(angles.shape[0]))
        kwargs.update(dict(blocksize=angles.shape[0]))
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)


sirt = decorator(SIRT, name="sirt")


class OS_SART(IterativeReconAlg):  
    __doc__ = (
        "OS_SART solves Cone Beam CT image reconstruction using Oriented Subsets\n"
        "Simultaneous Algebraic Reconxtruction Techique algorithm\n"
        "OS_SART(PROJ,GEO,ALPHA,NITER,BLOCKSIZE=20) solves the reconstruction problem\n"
        "using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "to the geometry descrived in GEO, using NITER iterations.\n"
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        
        self.blocksize = 20 if 'blocksize' not in kwargs else kwargs["blocksize"]       
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)


ossart = decorator(OS_SART, name="ossart")


class SART_TV(IterativeReconAlg):  
    __doc__ = (
        "SART_TV solves Cone Beam CT image reconstruction using Simultaneous \n"
        "Algebraic Reconstruction Technique with TV regularization algorithm\n"
        "SART_TV(PROJ,GEO,ALPHA,NITER,TVLAMBDA=50,TVITER=50) solves the reconstruction\n"
        "problem using the projection data PROJ taken over ALPHA angles\n"
        "corresponding to the geometry described in GEO, using NITER iterations. \n"
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        
        if "blocksize" in kwargs and kwargs['blocksize']>1:
            print('Warning: blocksize is set to 1, please use an OS version of the algorithm for blocksize > 1')
        kwargs.update(dict(blocksize=1))
        self.tvlambda = 50 if 'tvlambda' not in kwargs else kwargs['tvlambda']
        self.tviter = 50 if 'tviter' not in kwargs else kwargs['tviter']
        # these two settings work well for nVoxel=[254,254,254]

        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

    # Override
    def run_main_iter(self):
        """
        Goes through the main iteration for the given configuration.
        :return: None
        """
        Quameasopts = self.Quameasopts

        for i in range(self.niter):

            res_prev = None
            if Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)
            if self.verbose:
                self._estimate_time_until_completion(i)

            getattr(self, self.dataminimizing)()
            # print("run_main_iter: gpuids = {}", self.gpuids)
            self.res = im3ddenoise(self.res, self.tviter, self.tvlambda, self.gpuids)
            if Quameasopts is not None:
                self.error_measurement(res_prev, i)


sart_tv = decorator(SART_TV, name="sart_tv")


class OSSART_TV(IterativeReconAlg):  
    __doc__ = (
        "OSSART_TV solves Cone Beam CT image reconstruction using Oriented Subsets\n"
        "Simultaneous Algebraic Reconxtruction Technique with TV regularization algorithm\n"
        "OSSART_TV(PROJ,GEO,ALPHA,NITER,BLOCKSIZE=20,TVLAMBDA=50,TVITER=50) \n"
        "solves the reconstruction problem using the projection data PROJ taken\n"
        "over ALPHA angles, corresponding to the geometry descrived in GEO,\n"
        "using NITER iterations.\n"
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        
        self.blocksize = 20 if 'blocksize' not in kwargs else kwargs['blocksize']
        self.tvlambda = 50 if 'tvlambda' not in kwargs else kwargs['tvlambda']
        self.tviter = 50 if 'tviter' not in kwargs else kwargs['tviter']
        # these two settings work well for nVoxel=[254,254,254]

        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)


ossart_tv = decorator(OSSART_TV, name="ossart_tv")
